from typing import Literal, Optional

from pydantic import BaseModel

from galileo_core.schemas.core.integration.integration_type import LLMIntegration


class OpenAIExtras(BaseModel):
    organization_id: Optional[str] = None


class OpenAIIntegrationCreate(OpenAIExtras):
    name: Literal[LLMIntegration.openai] = LLMIntegration.openai
    token: str



from typing import Dict, List, Optional

from pydantic import BaseModel, Field

from galileo_core.schemas.protect.payload import Payload
from galileo_core.schemas.protect.response import Response
from galileo_core.schemas.shared.workflows.step import (
    AgentStep,
    AWorkflowStep,
    LlmStep,
    LlmStepAllowedIOType,
    RetrieverStep,
    RetrieverStepAllowedOutputType,
    StepIOType,
    StepWithChildren,
    ToolStep,
    WorkflowStep,
)


class Workflows(BaseModel):
    workflows: List[AWorkflowStep] = Field(default_factory=list, description="List of workflows.")
    current_workflow: Optional[StepWithChildren] = Field(default=None, description="Current workflow in the workflow.")

    def add_workflow(
        self,
        input: StepIOType,
        output: Optional[StepIOType] = None,
        redacted_input: Optional[StepIOType] = None,
        redacted_output: Optional[StepIOType] = None,
        name: Optional[str] = None,
        duration_ns: Optional[int] = None,
        created_at_ns: Optional[int] = None,
        metadata: Optional[Dict[str, str]] = None,
        ground_truth: Optional[str] = None,
    ) -> WorkflowStep:
        """
        Create a new workflow and add it to the list of workflows.
        Simple usage:
        ```
        my_workflows.add_workflow("input")
        my_workflows.add_llm_step("input", "output", model="<my_model>")
        my_workflows.conclude_workflow("output")
        ```
        Parameters:
        ----------
            input: StepIOType: Input to the node.
            output: Optional[str]: Output of the node.
            redacted_input: Optional[StepIOType]: Redacted input to the node.
            redacted_output: Optional[str]: Redacted output of the node.
            name: Optional[str]: Name of the workflow.
            duration_ns: Optional[int]: Duration of the workflow in nanoseconds.
            created_at_ns: Optional[int]: Timestamp of the workflow's creation.
            metadata: Optional[Dict[str, str]]: Metadata associated with this workflow.
            ground_truth: Optional[str]: Ground truth, expected output of the workflow.
        Returns:
        -------
            WorkflowStep: The created workflow.
        """
        workflow = WorkflowStep(
            input=input,
            output=output or "",
            redacted_input=redacted_input,
            redacted_output=redacted_output,
            name=name,
            duration_ns=duration_ns,
            created_at_ns=created_at_ns,
            metadata=metadata,
            ground_truth=ground_truth,
        )
        self.workflows.append(workflow)
        self.current_workflow = workflow
        return workflow

    def add_agent_workflow(
        self,
        input: StepIOType,
        output: Optional[StepIOType] = None,
        redacted_input: Optional[StepIOType] = None,
        redacted_output: Optional[StepIOType] = None,
        name: Optional[str] = None,
        duration_ns: Optional[int] = None,
        created_at_ns: Optional[int] = None,
        metadata: Optional[Dict[str, str]] = None,
        ground_truth: Optional[str] = None,
    ) -> AgentStep:
        """
        Create a new workflow and add it to the list of workflows.
        Simple usage:
        ```
        my_workflows.add_agent_workflow("input")
        my_workflows.add_tool_step("input", "output")
        my_workflows.conclude_workflow("output")
        Parameters:
        ----------
            input: StepIOType: Input to the node.
            output: Optional[str]: Output of the node.
            redacted_input: Optional[StepIOType]: Redacted input to the node.
            redacted_output: Optional[str]: Redacted output of the node.
            name: Optional[str]: Name of the workflow.
            duration_ns: Optional[int]: Duration of the workflow in nanoseconds.
            created_at_ns: Optional[int]: Timestamp of the workflow's creation.
            metadata: Optional[Dict[str, str]]: Metadata associated with this workflow.
            ground_truth: Optional[str] = None, Ground truth, expected output of the workflow.
        Returns:
        -------
            AgentStep: The created agent workflow.
        """
        workflow = AgentStep(
            input=input,
            output=output or "",
            redacted_input=redacted_input,
            redacted_output=redacted_output,
            name=name,
            duration_ns=duration_ns,
            created_at_ns=created_at_ns,
            metadata=metadata,
            ground_truth=ground_truth,
        )
        self.workflows.append(workflow)
        self.current_workflow = workflow
        return workflow

    def add_single_step_workflow(
        self,
        input: LlmStepAllowedIOType,
        output: LlmStepAllowedIOType,
        model: str,
        redacted_input: Optional[LlmStepAllowedIOType] = None,
        redacted_output: Optional[LlmStepAllowedIOType] = None,
        tools: Optional[List[Dict]] = None,
        name: Optional[str] = None,
        duration_ns: Optional[int] = None,
        created_at_ns: Optional[int] = None,
        metadata: Optional[Dict[str, str]] = None,
        input_tokens: Optional[int] = None,
        output_tokens: Optional[int] = None,
        total_tokens: Optional[int] = None,
        temperature: Optional[float] = None,
        time_to_first_token_ms: Optional[float] = None,
        ground_truth: Optional[str] = None,
        status_code: Optional[int] = None,
    ) -> LlmStep:
        """
        Create a new single-step workflow and add it to the list of workflows. This is just if you need a plain llm
        workflow with no surrounding steps.

        Parameters:
        ----------
            input: LlmStepAllowedIOType: Input to the node.
            output: LlmStepAllowedIOType: Output of the node.
            model: str: Model used for this step. Feedback from April: Good docs about what model names we use.
            redacted_input: Optional[LlmStepAllowedIOType]: Redacted input to the node.
            redacted_output: Optional[LlmStepAllowedIOType]: Redacted output of the node.
            tools: Optional[List[Dict]]: List of available tools passed to LLM on invocation.
            name: Optional[str]: Name of the step.
            duration_ns: Optional[int]: duration_ns of the node in nanoseconds.
            created_at_ns: Optional[int]: Timestamp of the step's creation.
            metadata: Optional[Dict[str, str]]: Metadata associated with this step.
            input_tokens: Optional[int]: Number of input tokens.
            output_tokens: Optional[int]: Number of output tokens.
            total_tokens: Optional[int]: Total number of tokens.
            temperature: Optional[float]: Temperature used for generation.
            time_to_first_token_ms: Optional[float]: Time taken to generate the first token.
            ground_truth: Optional[str]: Ground truth, expected output of the workflow.
            status_code: Optional[int]: Status code of the node execution.
        Returns:
        -------
            LlmStep: The created step.
        """
        step = LlmStep(
            input=input,
            output=output,
            redacted_input=redacted_input,
            redacted_output=redacted_output,
            model=model,
            tools=tools,
            name=name,
            duration_ns=duration_ns,
            created_at_ns=created_at_ns,
            metadata=metadata,
            input_tokens=input_tokens,
            output_tokens=output_tokens,
            total_tokens=total_tokens,
            temperature=temperature,
            time_to_first_token_ms=time_to_first_token_ms,
            status_code=status_code,
            ground_truth=ground_truth,
        )
        self.workflows.append(step)
        # Single step workflows are automatically concluded so we reset the current step.
        self.current_workflow = None
        return step

    def add_llm_step(
        self,
        input: LlmStepAllowedIOType,
        output: LlmStepAllowedIOType,
        model: str,
        redacted_input: Optional[LlmStepAllowedIOType] = None,
        redacted_output: Optional[LlmStepAllowedIOType] = None,
        tools: Optional[List[Dict]] = None,
        name: Optional[str] = None,
        duration_ns: Optional[int] = None,
        created_at_ns: Optional[int] = None,
        metadata: Optional[Dict[str, str]] = None,
        input_tokens: Optional[int] = None,
        output_tokens: Optional[int] = None,
        total_tokens: Optional[int] = None,
        temperature: Optional[float] = None,
        time_to_first_token_ms: Optional[float] = None,
        status_code: Optional[int] = None,
    ) -> LlmStep:
        """
        Add a new llm step to the current workflow.

        Parameters:
        ----------
            input: LlmStepAllowedIOType: Input to the node.
            output: LlmStepAllowedIOType: Output of the node.
            model: str: Model used for this step.
            redacted_input: Optional[LlmStepAllowedIOType]: Redacted input to the node.
            redacted_output: Optional[LlmStepAllowedIOType]: Redacted output of the node.
            tools: Optional[List[Dict]]: List of available tools passed to LLM on invocation.
            name: Optional[str]: Name of the step.
            duration_ns: Optional[int]: duration_ns of the node in nanoseconds.
            created_at_ns: Optional[int]: Timestamp of the step's creation.
            metadata: Optional[Dict[str, str]]: Metadata associated with this step.
            input_tokens: Optional[int]: Number of input tokens.
            output_tokens: Optional[int]: Number of output tokens.
            total_tokens: Optional[int]: Total number of tokens.
            temperature: Optional[float]: Temperature used for generation.
            time_to_first_token_ms: Optional[float]: Time taken to generate the first token.
            status_code: Optional[int]: Status code of the node execution.
        Returns:
        -------
            LlmStep: The created step.
        """
        if self.current_workflow is None:
            raise ValueError("A workflow needs to be created in order to add a step.")
        step = self.current_workflow.add_llm(
            input=input,
            output=output,
            redacted_input=redacted_input,
            redacted_output=redacted_output,
            model=model,
            tools=tools,
            name=name,
            duration_ns=duration_ns,
            created_at_ns=created_at_ns,
            metadata=metadata,
            input_tokens=input_tokens,
            output_tokens=output_tokens,
            total_tokens=total_tokens,
            temperature=temperature,
            time_to_first_token_ms=time_to_first_token_ms,
            status_code=status_code,
        )
        return step

    def add_retriever_step(
        self,
        input: StepIOType,
        documents: RetrieverStepAllowedOutputType,
        redacted_input: Optional[StepIOType] = None,
        redacted_documents: Optional[RetrieverStepAllowedOutputType] = None,
        name: Optional[str] = None,
        duration_ns: Optional[int] = None,
        created_at_ns: Optional[int] = None,
        metadata: Optional[Dict[str, str]] = None,
        status_code: Optional[int] = None,
    ) -> RetrieverStep:
        """
        Add a new retriever step to the current workflow.

        Parameters:
        ----------
            input: StepIOType: Input to the node.
            documents: Union[List[str], List[Dict[str, str]], List[Document]]: Documents retrieved from the retriever.
            redacted_input: Optional[StepIOType]: Redacted input to the node.
            redacted_documents: Optional[RetrieverStepAllowedOutputType]: Redacted documents retrieved from the retriever.
            name: Optional[str]: Name of the step.
            duration_ns: Optional[int]: duration_ns of the node in nanoseconds.
            created_at_ns: Optional[int]: Timestamp of the step's creation.
            metadata: Optional[Dict[str, str]]: Metadata associated with this step.
            status_code: Optional[int]: Status code of the node execution.
        Returns:
        -------
            RetrieverStep: The created step.
        """
        if self.current_workflow is None:
            raise ValueError("A workflow needs to be created in order to add a step.")
        step = self.current_workflow.add_retriever(
            input=input,
            documents=documents,
            redacted_input=redacted_input,
            redacted_documents=redacted_documents,
            name=name,
            duration_ns=duration_ns,
            created_at_ns=created_at_ns,
            metadata=metadata,
            status_code=status_code,
        )
        return step

    def add_tool_step(
        self,
        input: StepIOType,
        output: StepIOType,
        redacted_input: Optional[StepIOType] = None,
        redacted_output: Optional[StepIOType] = None,
        name: Optional[str] = None,
        duration_ns: Optional[int] = None,
        created_at_ns: Optional[int] = None,
        metadata: Optional[Dict[str, str]] = None,
        status_code: Optional[int] = None,
    ) -> ToolStep:
        """
        Add a new tool step to the current workflow.

        Parameters:
        ----------
            input: StepIOType: Input to the node.
            output: StepIOType: Output of the node.
            redacted_input: Optional[StepIOType]: Redacted input to the node.
            redacted_output: Optional[StepIOType]: Redacted output of the node.
            name: Optional[str]: Name of the step.
            duration_ns: Optional[int]: duration_ns of the node in nanoseconds.
            created_at_ns: Optional[int]: Timestamp of the step's creation.
            metadata: Optional[Dict[str, str]]: Metadata associated with this step.
            status_code: Optional[int]: Status code of the node execution.
        Returns:
        -------
            ToolStep: The created step.
        """
        if self.current_workflow is None:
            raise ValueError("A workflow needs to be created in order to add a step.")
        step = self.current_workflow.add_tool(
            input=input,
            output=output,
            redacted_input=redacted_input,
            redacted_output=redacted_output,
            name=name,
            duration_ns=duration_ns,
            created_at_ns=created_at_ns,
            metadata=metadata,
            status_code=status_code,
        )
        return step

    def add_protect_step(
        self,
        payload: Payload,
        response: Response,
        redacted_payload: Optional[Payload] = None,
        redacted_response: Optional[Response] = None,
        duration_ns: Optional[int] = None,
        created_at_ns: Optional[int] = None,
        metadata: Optional[Dict[str, str]] = None,
        status_code: Optional[int] = None,
    ) -> ToolStep:
        """
        Add a new protect step to the current workflow.

        Parameters:
        ----------
            payload: Payload: Input to Protect `invoke`.
            response: Response: Output from Protect `invoke`.
            redacted_payload: Optional[Payload]: Redacted input to Protect `invoke`.
            redacted_response: Optional[Response]: Redacted output from Protect `invoke`.
            duration_ns: Optional[int]: duration_ns of the node in nanoseconds.
            created_at_ns: Optional[int]: Timestamp of the step's creation.
            metadata: Optional[Dict[str, str]]: Metadata associated with this step.
            status_code: Optional[int]: Status code of the node execution.
        Returns:
        -------
            ToolStep: The created step.
        """
        if self.current_workflow is None:
            raise ValueError("A workflow needs to be created in order to add a step.")
        step = self.current_workflow.add_protect(
            payload=payload,
            response=response,
            redacted_payload=redacted_payload,
            redacted_response=redacted_response,
            created_at_ns=created_at_ns,
            metadata=metadata,
            status_code=status_code,
        )
        return step

    def add_workflow_step(
        self,
        input: StepIOType,
        output: Optional[StepIOType] = None,
        redacted_input: Optional[StepIOType] = None,
        redacted_output: Optional[StepIOType] = None,
        name: Optional[str] = None,
        duration_ns: Optional[int] = None,
        created_at_ns: Optional[int] = None,
        metadata: Optional[Dict[str, str]] = None,
    ) -> WorkflowStep:
        """
        Add a nested workflow step to the workflow. This is useful when you want to create a nested workflow within the
        current workflow. The next step you add will be a child of this workflow. To step out of the nested workflow,
        use conclude_workflow().

        Parameters:
        ----------
            input: StepIOType: Input to the node.
            output: Optional[StepIOType]: Output of the node. This can also be set on conclude_workflow().
            redacted_input: Optional[StepIOType]: Redacted input to the node.
            redacted_output: Optional[StepIOType]: Redacted output of the node.
            name: Optional[str]: Name of the step.
            duration_ns: Optional[int]: duration_ns of the node in nanoseconds.
            created_at_ns: Optional[int]: Timestamp of the step's creation.
            metadata: Optional[Dict[str, str]]: Metadata associated with this step.
        Returns:
        -------
            WorkflowStep: The created step.
        """
        if self.current_workflow is None:
            raise ValueError("A workflow needs to be created in order to add a step.")
        step = self.current_workflow.add_sub_workflow(
            input=input,
            output=output,
            redacted_input=redacted_input,
            redacted_output=redacted_output,
            name=name,
            duration_ns=duration_ns,
            created_at_ns=created_at_ns,
            metadata=metadata,
        )
        self.current_workflow = step
        return step

    def add_agent_step(
        self,
        input: StepIOType,
        output: Optional[StepIOType] = None,
        redacted_input: Optional[StepIOType] = None,
        redacted_output: Optional[StepIOType] = None,
        name: Optional[str] = None,
        duration_ns: Optional[int] = None,
        created_at_ns: Optional[int] = None,
        metadata: Optional[Dict[str, str]] = None,
    ) -> AgentStep:
        """
        Add a nested agent workflow step to the workflow. This is useful when you want to create a nested workflow
        within the current workflow. The next step you add will be a child of this workflow. To step out of the nested
        workflow, use conclude_workflow().

        Parameters:
        ----------
            input: StepIOType: Input to the node.
            output: Optional[StepIOType]: Output of the node. This can also be set on conclude_workflow().
            redacted_input: Optional[StepIOType]: Redacted input to the node.
            redacted_output: Optional[StepIOType]: Redacted output of the node.
            name: Optional[str]: Name of the step.
            duration_ns: Optional[int]: duration_ns of the node in nanoseconds.
            created_at_ns: Optional[int]: Timestamp of the step's creation.
            metadata: Optional[Dict[str, str]]: Metadata associated with this step.
        Returns:
        -------
            AgentStep: The created step.
        """
        if self.current_workflow is None:
            raise ValueError("A workflow needs to be created in order to add a step.")
        step = self.current_workflow.add_sub_agent(
            input=input,
            output=output,
            redacted_input=redacted_input,
            redacted_output=redacted_output,
            name=name,
            duration_ns=duration_ns,
            created_at_ns=created_at_ns,
            metadata=metadata,
        )
        self.current_workflow = step
        return step

    def conclude_workflow(
        self,
        output: Optional[StepIOType] = None,
        redacted_output: Optional[StepIOType] = None,
        duration_ns: Optional[int] = None,
        status_code: Optional[int] = None,
    ) -> Optional[StepWithChildren]:
        """
        Conclude the workflow by setting the output of the current node. In the case of nested workflows, this will
        point the workflow back to the parent of the current workflow.

        Parameters:
        ----------
            output: Optional[StepIOType]: Output of the node.
            redacted_output: Optional[StepIOType]: Redacted output of the node.
            duration_ns: Optional[int]: duration_ns of the node in nanoseconds.
            status_code: Optional[int]: Status code of the node execution.
        Returns:
        -------
            Optional[StepWithChildren]: The parent of the current workflow. None if no parent exists.
        """
        if self.current_workflow is None:
            raise ValueError("No existing workflow to conclude.")
        self.current_workflow = self.current_workflow.conclude(
            output=output, redacted_output=redacted_output, duration_ns=duration_ns, status_code=status_code
        )
        return self.current_workflow



from typing import Union

from pydantic import Field, TypeAdapter
from typing_extensions import Annotated

from galileo_core.schemas.logging.session import Session
from galileo_core.schemas.logging.span import Span
from galileo_core.schemas.logging.trace import Trace

Step = Annotated[Union[Session, Trace, Span], Field(discriminator="type")]

StepAdapter: TypeAdapter[Step] = TypeAdapter(Step)



from enum import Enum


class AzureAuthenticationType(str, Enum):
    api_key = "api_key"
    client_secret = "client_secret"
    username_password = "username_password"
    custom_oauth2 = "custom_oauth2"



from json import JSONEncoder
from typing import Any

from pydantic import BaseModel


class PydanticJsonEncoder(JSONEncoder):
    def default(self, obj: Any) -> Any:
        if isinstance(obj, BaseModel):
            return obj.model_dump(mode="json", exclude_none=True, exclude_unset=True)
        return super().default(obj)



from typing import List, Optional

from pydantic import UUID4

from galileo_core.constants.request_method import RequestMethod
from galileo_core.constants.routes import Routes
from galileo_core.helpers.logger import logger
from galileo_core.schemas.base_config import GalileoConfig
from galileo_core.schemas.core.metric_critique import (
    CreateMetricCritiquesRequest,
    MetricCritique,
    MetricCritiqueResponse,
)


def create_metric_critiques(
    project_id: UUID4,
    metric_name: str,
    critiques: List[MetricCritique],
    config: Optional[GalileoConfig] = None,
) -> List[MetricCritiqueResponse]:
    config = config or GalileoConfig.get()

    logger.debug(
        "Creating metric critiques.",
        extra=dict(project_id=project_id, metric_name=metric_name, n_critiques=len(critiques)),
    )

    request = CreateMetricCritiquesRequest.model_validate(
        dict(metric=metric_name, critiques=[critique.model_dump() for critique in critiques])
    )

    response_list = config.api_client.request(
        RequestMethod.POST,
        Routes.metric_critiques.format(project_id=project_id),
        json=request.model_dump(mode="json"),
    )

    return [MetricCritiqueResponse.model_validate(response_dict) for response_dict in response_list]



import math
from datetime import datetime, timedelta, timezone
from json import dumps
from typing import Any, Dict, List, Literal, Optional, Sequence, Union, cast

from pydantic import ConfigDict, Field, PrivateAttr, TypeAdapter, field_validator
from pydantic.types import UUID4
from pydantic_core import ValidationError
from typing_extensions import Annotated

from galileo_core.schemas.logging.agent import AgentType
from galileo_core.schemas.logging.llm import Event, Message, MessageRole
from galileo_core.schemas.logging.step import BaseStep, Metrics, StepAllowedInputType, StepType
from galileo_core.schemas.shared.document import Document
from galileo_core.utils.json import PydanticJsonEncoder

LlmSpanAllowedInputType = Union[Sequence[Message], Message, str, Dict[str, Any], Sequence[Dict[str, Any]]]
LlmSpanAllowedOutputType = Union[Message, str, Dict[str, Any]]


class BaseSpan(BaseStep):
    input: StepAllowedInputType = Field(default="", description="Input to the trace or span.")
    redacted_input: Optional[StepAllowedInputType] = Field(
        default=None, description="Redacted input of the trace or span."
    )
    step_number: Optional[int] = Field(default=None, description="Topological step number of the span.")
    parent_id: Optional[UUID4] = Field(
        default=None,
        title="Parent ID",
        description="Galileo ID of the parent of this span",
    )


class StepWithChildSpans(BaseSpan):
    spans: List["Span"] = Field(default_factory=list, description="Child spans.")
    _last_child_created_at: Optional[datetime] = PrivateAttr(default=None)

    def add_child_spans(self, spans: Sequence["Span"]) -> None:
        self.spans.extend(spans)

    def add_child_span(self, span: "Span") -> None:
        self.add_child_spans([span])

    def _generate_unique_timestamp(self) -> datetime:
        """
        Generate a unique timestamp for a new child step, ensuring it is always greater than the previous one.

        This is important for spans that are created in rapid succession, as it prevents them from having the same
        timestamp, which can cause ordering issues in the UI.
        """
        new_timestamp = datetime.now(timezone.utc)
        if self._last_child_created_at is not None and new_timestamp <= self._last_child_created_at:
            new_timestamp = self._last_child_created_at + timedelta(microseconds=1)

        self._last_child_created_at = new_timestamp
        return new_timestamp


class BaseWorkflowSpan(BaseSpan):
    type: Literal[StepType.workflow] = Field(
        default=StepType.workflow, description=BaseStep.model_fields["type"].description
    )


class WorkflowSpan(BaseWorkflowSpan, StepWithChildSpans):
    pass


class BaseAgentSpan(BaseSpan):
    type: Literal[StepType.agent] = Field(default=StepType.agent, description=BaseStep.model_fields["type"].description)
    agent_type: AgentType = Field(default=AgentType.default, description="Agent type.")


class AgentSpan(BaseAgentSpan, StepWithChildSpans):
    pass


class LlmMetrics(Metrics):
    num_input_tokens: Optional[int] = Field(default=None, description="Number of input tokens.")
    num_output_tokens: Optional[int] = Field(default=None, description="Number of output tokens.")
    num_total_tokens: Optional[int] = Field(default=None, description="Total number of tokens.")
    time_to_first_token_ns: Optional[int] = Field(
        default=None,
        description="Time until the first token was generated in nanoseconds.",
    )

    model_config = ConfigDict(extra="allow")


class LlmSpan(BaseSpan):
    type: Literal[StepType.llm] = Field(default=StepType.llm, description=BaseStep.model_fields["type"].description)
    input: Sequence[Message] = Field(
        default_factory=list, validate_default=True, description=BaseStep.model_fields["input"].description
    )
    redacted_input: Optional[Sequence[Message]] = Field(
        default=None, description=BaseStep.model_fields["redacted_input"].description
    )
    output: Message = Field(
        default_factory=lambda: Message(content="", role=MessageRole.assistant),
        validate_default=True,
        description=BaseStep.model_fields["output"].description,
    )
    redacted_output: Optional[Message] = Field(
        default=None, description=BaseStep.model_fields["redacted_output"].description
    )
    metrics: LlmMetrics = Field(default_factory=LlmMetrics, description=BaseStep.model_fields["metrics"].description)
    tools: Optional[Sequence[Dict[str, Any]]] = Field(
        default=None,
        description="List of available tools passed to the LLM on invocation.",
    )
    events: Optional[List[Event]] = Field(
        default=None,
        description="List of reasoning, internal tool call, or MCP events that occurred during the LLM span.",
    )
    model: Optional[str] = Field(default=None, description="Model used for this span.")
    temperature: Optional[float] = Field(default=None, description="Temperature used for generation.")
    finish_reason: Optional[str] = Field(default=None, description="Reason for finishing.")

    @classmethod
    def _convert_dict_to_message(cls, value: Dict[str, Any], default_role: MessageRole = MessageRole.user) -> Message:
        """
        Converts a dict into a Message object.
        Will dump the dict to a json string if it unable to be deserialized into a Message object.

        Args:
            value (Dict[str, Any]): The dict to convert.
            default_role (Optional[MessageRole], optional): The role to use if the dict does not contain a role. Defaults to MessageRole.user.

        Returns:
            Message: The converted Message object.
        """
        try:
            return Message.model_validate(value)
        except ValidationError:
            return Message(content=dumps(value), role=default_role)

    @field_validator("tools", mode="after")
    def validate_tools_serializable(cls, val: Optional[Sequence[Dict[str, Any]]]) -> Optional[Sequence[Dict[str, Any]]]:
        # Make sure we can dump input/output to json string.
        dumps(val, cls=PydanticJsonEncoder)
        return val

    @classmethod
    def _convert_input_to_messages(cls, value: LlmSpanAllowedInputType) -> Sequence[Message]:
        """Helper method to convert various input types into a standardized list of Message objects."""
        if isinstance(value, Sequence) and all(isinstance(item, Message) for item in value):
            return cast(Sequence[Message], value)
        if isinstance(value, Sequence) and all(isinstance(item, Dict) for item in value):
            return [
                cls._convert_dict_to_message(value=cast(Dict[str, Any], item), default_role=MessageRole.user)
                for item in value
            ]
        if isinstance(value, str):
            return [Message(role=MessageRole.user, content=value)]
        if isinstance(value, Message):
            return [value]
        if isinstance(value, Dict):
            return [cls._convert_dict_to_message(value=value, default_role=MessageRole.user)]
        raise ValueError("LLM span input must be a Message, a list of Messages, a dict, a list of dicts, or a string.")

    @classmethod
    def _convert_output_to_message(cls, value: LlmSpanAllowedOutputType) -> Message:
        """Helper method to convert various output types into a standardized Message object."""
        if isinstance(value, Message):
            return value
        if isinstance(value, str):
            return Message(role=MessageRole.assistant, content=value)
        if isinstance(value, Dict):
            return cls._convert_dict_to_message(value=value, default_role=MessageRole.assistant)
        raise ValueError("LLM span output must be a Message, a string, or a dict.")

    @field_validator("input", mode="before")
    def convert_input(cls, value: LlmSpanAllowedInputType) -> Sequence[Message]:
        """Converts various input types into a standardized list of Message objects."""
        return cls._convert_input_to_messages(value)

    @field_validator("redacted_input", mode="before")
    def convert_redacted_input(cls, value: Optional[LlmSpanAllowedInputType]) -> Optional[Sequence[Message]]:
        """Converts various redacted input types into a standardized list of Message objects."""
        if value is None:
            return None
        return cls._convert_input_to_messages(value)

    @field_validator("output", mode="before")
    def convert_output(cls, value: LlmSpanAllowedOutputType) -> Message:
        """Converts various output types into a standardized Message object."""
        return cls._convert_output_to_message(value)

    @field_validator("redacted_output", mode="before")
    def convert_redacted_output(cls, value: Optional[LlmSpanAllowedOutputType]) -> Optional[Message]:
        """Converts various redacted output types into a standardized Message object."""
        if value is None:
            return None
        return cls._convert_output_to_message(value)

    @field_validator("temperature", mode="before")
    def convert_temperature(cls, value: Optional[float]) -> Optional[float]:
        if value is None or math.isnan(value) or math.isinf(value):
            return None
        return value


class BaseRetrieverSpan(BaseSpan):
    type: Literal[StepType.retriever] = Field(
        default=StepType.retriever, description=BaseStep.model_fields["type"].description
    )
    input: str = Field(default="", validate_default=True, description=BaseStep.model_fields["input"].description)
    redacted_input: Optional[str] = Field(default=None, description=BaseStep.model_fields["redacted_input"].description)
    output: List[Document] = Field(
        default_factory=list, validate_default=True, description=BaseStep.model_fields["output"].description
    )
    redacted_output: Optional[List[Document]] = Field(
        default=None, description=BaseStep.model_fields["redacted_output"].description
    )

    @classmethod
    def _convert_to_documents(cls, value: Union[List[Dict[str, str]], List[Document]]) -> List[Document]:
        """Helper method to convert various output types into a standardized list of Document objects."""
        if not isinstance(value, list):
            raise ValueError("Retriever output must be a list of dicts or a list of Documents.")

        if all(isinstance(doc, Document) for doc in value):
            return cast(List[Document], value)

        if all(isinstance(doc, dict) for doc in value):
            allowed_fields = {"content", "page_content", "metadata"}
            filtered_docs = [
                {k: v for k, v in cast(Dict[str, Any], doc).items() if k in allowed_fields} for doc in value
            ]
            return [Document.model_validate(doc) for doc in filtered_docs]

        raise ValueError("Retriever output must be a list of dicts, or a list of Documents.")

    @field_validator("output", mode="before")
    def set_output(cls, value: Union[List[Dict[str, str]], List[Document]]) -> List[Document]:
        return cls._convert_to_documents(value)

    @field_validator("redacted_output", mode="before")
    def set_redacted_output(
        cls, value: Optional[Union[List[Dict[str, str]], List[Document]]]
    ) -> Optional[List[Document]]:
        if value is None:
            return None
        return cls._convert_to_documents(value)


class RetrieverSpan(BaseRetrieverSpan, StepWithChildSpans):
    pass


class BaseToolSpan(BaseSpan):
    type: Literal[StepType.tool] = Field(default=StepType.tool, description=BaseStep.model_fields["type"].description)
    input: str = Field(default="", description=BaseStep.model_fields["input"].description)
    redacted_input: Optional[str] = Field(default=None, description=BaseStep.model_fields["redacted_input"].description)
    output: Optional[str] = Field(default=None, description=BaseStep.model_fields["output"].description)
    redacted_output: Optional[str] = Field(
        default=None, description=BaseStep.model_fields["redacted_output"].description
    )
    tool_call_id: Optional[str] = Field(default=None, description="ID of the tool call.")


class ToolSpan(BaseToolSpan, StepWithChildSpans):
    pass


Span = Annotated[Union[AgentSpan, WorkflowSpan, LlmSpan, RetrieverSpan, ToolSpan], Field(discriminator="type")]

SpanAdapter: TypeAdapter[Span] = TypeAdapter(Span)

StepWithChildSpans.model_rebuild()

SpanStepTypes = [step_type.value for step_type in StepType if step_type != StepType.trace]



from enum import Enum
from typing import Literal


class CodeScorerLanguage(str, Enum):
    """Supported sandbox languages."""

    python = "python"
    javascript = "javascript"
    typescript = "typescript"

    @property
    def file_extension(self) -> str:
        """Get the executable file extension for this language."""
        extensions = {
            CodeScorerLanguage.python: ".py",
            CodeScorerLanguage.javascript: ".js",
            CodeScorerLanguage.typescript: ".ts",
        }
        return extensions.get(self, "")  # Default to nothing if language isn't recognized

    @classmethod
    def from_file_extension(cls, ext: str) -> "CodeScorerLanguage":
        """Infer the language from the file extension."""
        for lang in cls:
            if lang.file_extension == ext:
                return lang
        raise ValueError(f"Unsupported file extension: {ext}")

    @property
    def default_version(self) -> str:
        """Get the default version of the language."""
        versions = {
            CodeScorerLanguage.python: "3.12",
            CodeScorerLanguage.javascript: "18",
            CodeScorerLanguage.typescript: "5.0",
        }
        return versions.get(self, "")


class ScorerType(str, Enum):
    luna = "luna"
    plus = "plus"


PlusScorerType = Literal[ScorerType.plus]
LunaOrPlusScorerType = Literal[ScorerType.luna, ScorerType.plus]



from enum import Enum


class Scorers(str, Enum):
    completeness_luna = "completeness_nli"
    completeness_plus = "completeness_gpt"
    context_adherence_luna = "adherence_nli"
    context_adherence_plus = "groundedness"
    context_relevance = "context_relevance"
    correctness = "factuality"
    chunk_attribution_utilization_luna = "chunk_attribution_utilization_nli"
    chunk_attribution_utilization_plus = "chunk_attribution_utilization_gpt"
    pii = "pii"
    prompt_injection = "prompt_injection"
    prompt_injection_plus = "prompt_injection_gpt"
    prompt_perplexity = "prompt_perplexity"
    input_sexist = "input_sexist"
    input_sexist_plus = "input_sexist_gpt"
    sexist = "sexist"
    sexist_plus = "sexist_gpt"
    tone = "tone"
    input_toxicity = "input_toxicity"
    input_toxicity_plus = "input_toxicity_gpt"
    toxicity = "toxicity"
    toxicity_plus = "toxicity_gpt"
    instruction_adherence_plus = "instruction_adherence"
    ground_truth_adherence_plus = "ground_truth_adherence"
    tool_errors_plus = "tool_error_rate"
    tool_selection_quality_plus = "tool_selection_quality"
    action_advancement_plus = "agentic_workflow_success"
    action_completion_plus = "agentic_session_success"



from typing import List, Optional, Union

from annotated_types import Len
from pydantic import UUID4, BaseModel, ConfigDict, Field
from typing_extensions import Annotated


class MetricCritique(BaseModel):
    run_id: Optional[UUID4] = None  # Evaluate requires a run id, while observe does not.
    row_id: Union[UUID4, int]
    critique: str
    intended_value: bool = Field(
        description="Set to opposite of actual value for new critiques, and set as same as existing critique intended value for updates."
    )


class CreateMetricCritiquesRequest(BaseModel):
    critiques: Annotated[List[MetricCritique], Len(min_length=1)]
    metric: str


class MetricCritiqueResponse(BaseModel):
    id: UUID4
    project_id: UUID4
    metric: str
    scorer_id: UUID4
    row_id: Union[UUID4, int]
    is_computed: bool
    model_config = ConfigDict(extra="allow")



from functools import cached_property
from itertools import chain
from typing import List

from pydantic import BaseModel, Field, field_validator
from pydantic_core import Url

from galileo_core.helpers.logger import get_logger
from galileo_core.schemas.protect.execution_status import ExecutionStatus

logger = get_logger()


class SubscriptionConfig(BaseModel):
    statuses: List[ExecutionStatus] = Field(
        description="List of statuses that will cause a notification to be sent to the configured URL.",
        default=[ExecutionStatus.triggered],
    )
    url: Url = Field(
        description="URL to send the event to. This can be a webhook URL, a message queue URL, an event bus or "
        "a custom endpoint that can receive an HTTP POST request."
    )

    @cached_property
    def possible_statuses(self) -> List[str]:
        """
        List of possible statuses that will trigger a notification to be sent.

        This is a list of the cased and uncased versions of the statuses in the `statuses` list.

        Returns
        -------
        List[str]
            _description_
        """
        return list(chain.from_iterable([status, status.lower(), status.upper()] for status in self.statuses))

    @field_validator("statuses", mode="before")
    def warn_if_paused(cls, statuses: List[ExecutionStatus]) -> List[ExecutionStatus]:
        if ExecutionStatus.paused in statuses:
            logger.warning(
                "Only stages can have a `paused` status, rulesets cannot. The 'paused' status will be ignored."
            )
        return statuses



from collections import deque
from datetime import datetime
from json import dumps
from typing import Deque, Dict, List, Optional

from pydantic import BaseModel, Field
from pydantic.types import UUID4

from galileo_core.schemas.logging.agent import AgentType
from galileo_core.schemas.logging.llm import Event
from galileo_core.schemas.logging.span import (
    AgentSpan,
    LlmMetrics,
    LlmSpan,
    LlmSpanAllowedInputType,
    LlmSpanAllowedOutputType,
    RetrieverSpan,
    Span,
    StepWithChildSpans,
    ToolSpan,
    WorkflowSpan,
)
from galileo_core.schemas.logging.step import Metrics
from galileo_core.schemas.logging.trace import Trace
from galileo_core.schemas.shared.document import Document
from galileo_core.utils.json import PydanticJsonEncoder


class TracesLogger(BaseModel):
    traces: List[Trace] = Field(default_factory=list, description="List of traces.")
    _parent_stack: Deque[StepWithChildSpans] = deque()

    def current_parent(self) -> Optional[StepWithChildSpans]:
        return self._parent_stack[-1] if self._parent_stack else None

    def add_child_span_to_parent(self, span: Span) -> None:
        current_parent = self.current_parent()
        if current_parent is None:
            raise ValueError("A trace needs to be created in order to add a span.")
        span.dataset_input = current_parent.dataset_input
        span.dataset_output = current_parent.dataset_output
        span.dataset_metadata = current_parent.dataset_metadata
        current_parent.add_child_span(span)

    def _get_child_span_timestamp(self) -> datetime:
        """Gets a unique timestamp for a child span from its parent."""
        parent = self.current_parent()
        if not parent:
            raise ValueError("A parent trace or span is required to generate a timestamp.")
        return parent._generate_unique_timestamp()

    def add_trace(
        self,
        input: str,
        redacted_input: Optional[str] = None,
        output: Optional[str] = None,
        redacted_output: Optional[str] = None,
        name: Optional[str] = None,
        created_at: Optional[datetime] = None,
        duration_ns: Optional[int] = None,
        user_metadata: Optional[Dict[str, str]] = None,
        tags: Optional[List[str]] = None,
        dataset_input: Optional[str] = None,
        dataset_output: Optional[str] = None,
        dataset_metadata: Optional[Dict[str, str]] = None,
        external_id: Optional[str] = None,
        id: Optional[UUID4] = None,
    ) -> Trace:
        """
        Create a new trace and add it to the list of traces.
        Simple usage:
        ```
        my_traces.add_trace("input")
        my_traces.add_llm_span("input", "output", model="<my_model>")
        my_traces.conclude("output")
        ```
        Parameters:
        ----------
            input: str: Input to the node.
            redacted_input: Optional[str]: Redacted input to the node.
            output: Optional[str]: Output of the node.
            redacted_output: Optional[str]: Redacted output of the node.
            name: Optional[str]: Name of the trace.
            duration_ns: Optional[int]: Duration of the trace in nanoseconds.
            created_at: Optional[datetime]: Timestamp of the trace's creation.
            user_metadata: Optional[Dict[str, str]]: Metadata associated with this trace.
            ground_truth: Optional[str]: Ground truth, expected output of the trace.
        Returns:
        -------
            Trace: The created trace.
        """
        if self.current_parent() is not None:
            raise ValueError("You must conclude the existing trace before adding a new one.")
        trace = Trace(
            input=input,
            redacted_input=redacted_input,
            output=output,
            redacted_output=redacted_output,
            name=name,
            created_at=created_at,
            user_metadata=user_metadata,
            tags=tags,
            metrics=Metrics(duration_ns=duration_ns),
            dataset_input=dataset_input,
            dataset_output=dataset_output,
            dataset_metadata=dataset_metadata if dataset_metadata is not None else {},
            external_id=external_id,
            id=id,
        )
        self.traces.append(trace)
        self._parent_stack.append(trace)
        return trace

    def add_single_llm_span_trace(
        self,
        input: LlmSpanAllowedInputType,
        output: LlmSpanAllowedOutputType,
        model: Optional[str],
        redacted_input: Optional[LlmSpanAllowedInputType] = None,
        redacted_output: Optional[LlmSpanAllowedOutputType] = None,
        tools: Optional[List[Dict]] = None,
        events: Optional[List[Event]] = None,
        name: Optional[str] = None,
        created_at: Optional[datetime] = None,
        duration_ns: Optional[int] = None,
        user_metadata: Optional[Dict[str, str]] = None,
        tags: Optional[List[str]] = None,
        num_input_tokens: Optional[int] = None,
        num_output_tokens: Optional[int] = None,
        total_tokens: Optional[int] = None,
        temperature: Optional[float] = None,
        status_code: Optional[int] = None,
        time_to_first_token_ns: Optional[int] = None,
        dataset_input: Optional[str] = None,
        dataset_output: Optional[str] = None,
        dataset_metadata: Optional[Dict[str, str]] = None,
        trace_id: Optional[UUID4] = None,
        span_id: Optional[UUID4] = None,
        span_step_number: Optional[int] = None,
    ) -> Trace:
        """
        Create a new trace with a single span and add it to the list of traces.

        Parameters:
        ----------
            input: LlmStepAllowedIOType: Input to the node.
            output: LlmStepAllowedIOType: Output of the node.
            model: str: Model used for this span. Feedback from April: Good docs about what model names we use.
            redacted_input: Optional[LlmStepAllowedIOType]: Redacted input to the node.
            redacted_output: Optional[LlmStepAllowedIOType]: Redacted output of the node
            tools: Optional[List[Dict]]: List of available tools passed to LLM on invocation.
            events: Optional[List[Event]]: List of reasoning, internal tool call, or MCP events.
            name: Optional[str]: Name of the span.
            duration_ns: Optional[int]: duration_ns of the node in nanoseconds.
            created_at: Optional[datetime]: Timestamp of the span's creation.
            user_metadata: Optional[Dict[str, str]]: Metadata associated with this span.
            num_input_tokens: Optional[int]: Number of input tokens.
            num_output_tokens: Optional[int]: Number of output tokens.
            total_tokens: Optional[int]: Total number of tokens.
            temperature: Optional[float]: Temperature used for generation.
            ground_truth: Optional[str]: Ground truth, expected output of the workflow.
            status_code: Optional[int]: Status code of the node execution.
            time_to_first_token_ns: Optional[int]: Time until the first token was returned.
            dataset_input: Optional[str]: Dataset input for the span.
            dataset_output: Optional[str]: Dataset output for the span.
            dataset_metadata: Optional[Dict[str, str]]: Dataset metadata for the span.
            trace_id: Optional[UUID4]: ID of the trace.
            span_id: Optional[UUID4]: ID of the span.
            span_step_number: Optional[int]: Step number of the span.
        Returns:
        -------
            Trace: The created trace.
        """
        if self.current_parent() is not None:
            raise ValueError("A trace cannot be created within a parent trace or span, it must always be the root.")

        trace = Trace(
            input=dumps(input, cls=PydanticJsonEncoder),
            redacted_input=dumps(redacted_input, cls=PydanticJsonEncoder) if redacted_input else None,
            output=dumps(output, cls=PydanticJsonEncoder),
            redacted_output=dumps(redacted_output, cls=PydanticJsonEncoder) if redacted_output else None,
            name=name,
            created_at=created_at,
            user_metadata=user_metadata,
            tags=tags,
            dataset_input=dataset_input,
            dataset_output=dataset_output,
            dataset_metadata=dataset_metadata if dataset_metadata is not None else {},
            id=trace_id,
        )
        trace.add_child_span(
            LlmSpan(
                name=name,
                created_at=created_at,
                user_metadata=user_metadata,
                tags=tags,
                input=input,
                redacted_input=redacted_input,
                output=output,
                redacted_output=redacted_output,
                metrics=LlmMetrics(
                    duration_ns=duration_ns,
                    num_input_tokens=num_input_tokens,
                    num_output_tokens=num_output_tokens,
                    num_total_tokens=total_tokens,
                    time_to_first_token_ns=time_to_first_token_ns,
                ),
                tools=tools,
                events=events,
                model=model,
                temperature=temperature,
                status_code=status_code,
                dataset_input=dataset_input,
                dataset_output=dataset_output,
                dataset_metadata=dataset_metadata if dataset_metadata is not None else {},
                id=span_id,
                step_number=span_step_number,
            )
        )
        self.traces.append(trace)
        # Single span traces are automatically concluded so we reset the current parent.
        self._parent_stack = deque()
        return trace

    def add_llm_span(
        self,
        input: LlmSpanAllowedInputType,
        output: LlmSpanAllowedOutputType,
        model: Optional[str],
        redacted_input: Optional[LlmSpanAllowedInputType] = None,
        redacted_output: Optional[LlmSpanAllowedOutputType] = None,
        tools: Optional[List[Dict]] = None,
        events: Optional[List[Event]] = None,
        name: Optional[str] = None,
        created_at: Optional[datetime] = None,
        duration_ns: Optional[int] = None,
        user_metadata: Optional[Dict[str, str]] = None,
        tags: Optional[List[str]] = None,
        num_input_tokens: Optional[int] = None,
        num_output_tokens: Optional[int] = None,
        total_tokens: Optional[int] = None,
        temperature: Optional[float] = None,
        status_code: Optional[int] = None,
        time_to_first_token_ns: Optional[int] = None,
        id: Optional[UUID4] = None,
        step_number: Optional[int] = None,
    ) -> LlmSpan:
        """
        Add a new llm span to the current parent.

        Parameters:
        ----------
            input: LlmStepAllowedIOType: Input to the node.
            output: LlmStepAllowedIOType: Output of the node.
            model: str: Model used for this span.
            redacted_input: Optional[LlmStepAllowedIOType]: Redacted input to the node.
            redacted_output: Optional[LlmStepAllowedIOType]: Redacted output of the node
            tools: Optional[List[Dict]]: List of available tools passed to LLM on invocation.
            events: Optional[List[Event]]: List of reasoning, internal tool call, or MCP events.
            name: Optional[str]: Name of the span.
            duration_ns: Optional[int]: duration_ns of the node in nanoseconds.
            created_at: Optional[datetime]: Timestamp of the span's creation.
            user_metadata: Optional[Dict[str, str]]: Metadata associated with this span.
            num_input_tokens: Optional[int]: Number of input tokens.
            num_output_tokens: Optional[int]: Number of output tokens.
            total_tokens: Optional[int]: Total number of tokens.
            temperature: Optional[float]: Temperature used for generation.
            status_code: Optional[int]: Status code of the node execution.
            time_to_first_token_ns: Optional[int]: Time until the first token was returned.
            id: Optional[UUID4]: ID of the span.
            step_number: Optional[int]: Step number of the span.
        Returns:
        -------
            LlmSpan: The created span.
        """

        span = LlmSpan(
            input=input,
            redacted_input=redacted_input,
            output=output,
            redacted_output=redacted_output,
            name=name,
            created_at=self._get_child_span_timestamp() if created_at is None else created_at,
            user_metadata=user_metadata,
            tags=tags,
            metrics=LlmMetrics(
                duration_ns=duration_ns,
                num_input_tokens=num_input_tokens,
                num_output_tokens=num_output_tokens,
                num_total_tokens=total_tokens,
                time_to_first_token_ns=time_to_first_token_ns,
            ),
            tools=tools,
            events=events,
            model=model,
            temperature=temperature,
            status_code=status_code,
            id=id,
            step_number=step_number,
        )
        self.add_child_span_to_parent(span)
        return span

    def add_retriever_span(
        self,
        input: str,
        documents: List[Document],
        redacted_input: Optional[str] = None,
        redacted_documents: Optional[List[Document]] = None,
        name: Optional[str] = None,
        duration_ns: Optional[int] = None,
        created_at: Optional[datetime] = None,
        user_metadata: Optional[Dict[str, str]] = None,
        tags: Optional[List[str]] = None,
        status_code: Optional[int] = None,
        id: Optional[UUID4] = None,
        step_number: Optional[int] = None,
    ) -> RetrieverSpan:
        """
        Add a new retriever span to the current parent.

        Parameters:
        ----------
            input: str: Input to the node.
            documents: List[Document]: Documents retrieved from the retriever.
            redacted_input: Optional[str]: Redacted input to the node.
            redacted_documents: Optional[List[Document]]: Redacted documents retrieved from the retriever.
            name: Optional[str]: Name of the span.
            duration_ns: Optional[int]: duration_ns of the node in nanoseconds.
            created_at: Optional[datetime]: Timestamp of the span's creation.
            user_metadata: Optional[Dict[str, str]]: Metadata associated with this span.
            status_code: Optional[int]: Status code of the node execution.
            id: Optional[UUID4]: ID of the span.
            step_number: Optional[int]: Step number of the span.
        Returns:
        -------
            RetrieverSpan: The created span.
        """
        span = RetrieverSpan(
            input=input,
            redacted_input=redacted_input,
            output=documents,
            redacted_output=redacted_documents,
            name=name,
            created_at=self._get_child_span_timestamp() if created_at is None else created_at,
            user_metadata=user_metadata,
            tags=tags,
            status_code=status_code,
            metrics=Metrics(duration_ns=duration_ns),
            id=id,
            step_number=step_number,
        )
        self.add_child_span_to_parent(span)
        return span

    def add_tool_span(
        self,
        input: str,
        redacted_input: Optional[str] = None,
        output: Optional[str] = None,
        redacted_output: Optional[str] = None,
        name: Optional[str] = None,
        duration_ns: Optional[int] = None,
        created_at: Optional[datetime] = None,
        user_metadata: Optional[Dict[str, str]] = None,
        tags: Optional[List[str]] = None,
        status_code: Optional[int] = None,
        tool_call_id: Optional[str] = None,
        id: Optional[UUID4] = None,
        step_number: Optional[int] = None,
    ) -> ToolSpan:
        """
        Add a new tool span to the current parent.

        Parameters:
        ----------
            input: str: Input to the node.
            redacted_input: Optional[str]: Redacted input to the node.
            output: Optional[str]: Output of the node.
            redacted_output: Optional[str]: Redacted output of the node.
            name: Optional[str]: Name of the span.
            duration_ns: Optional[int]: duration_ns of the node in nanoseconds.
            created_at: Optional[datetime]: Timestamp of the span's creation.
            user_metadata: Optional[Dict[str, str]]: Metadata associated with this span.
            status_code: Optional[int]: Status code of the node execution.
            tool_call_id: Optional[str]: ID of the tool call.
            id: Optional[UUID4]: ID of the span.
            step_number: Optional[int]: Step number of the span.
        Returns:
        -------
            ToolSpan: The created span.
        """
        span = ToolSpan(
            input=input,
            redacted_input=redacted_input,
            output=output,
            redacted_output=redacted_output,
            name=name,
            created_at=self._get_child_span_timestamp() if created_at is None else created_at,
            user_metadata=user_metadata,
            tags=tags,
            status_code=status_code,
            tool_call_id=tool_call_id,
            metrics=Metrics(duration_ns=duration_ns),
            id=id,
            step_number=step_number,
        )
        self.add_child_span_to_parent(span)
        return span

    def add_workflow_span(
        self,
        input: str,
        redacted_input: Optional[str] = None,
        output: Optional[str] = None,
        redacted_output: Optional[str] = None,
        name: Optional[str] = None,
        duration_ns: Optional[int] = None,
        created_at: Optional[datetime] = None,
        user_metadata: Optional[Dict[str, str]] = None,
        tags: Optional[List[str]] = None,
        id: Optional[UUID4] = None,
        step_number: Optional[int] = None,
    ) -> WorkflowSpan:
        """
        Add a workflow span to the current parent. This is useful when you want to create a nested workflow span
        within the trace or current workflow span. The next span you add will be a child of the current parent. To
        move out of the nested workflow, use conclude().

        Parameters:
        ----------
            input: str: Input to the node.
            redacted_input: Optional[str]: Redacted input to the node.
            output: Optional[str]: Output of the node. This can also be set on conclude().
            redacted_output: Optional[str]: Redacted output of the node.
            name: Optional[str]: Name of the span.
            duration_ns: Optional[int]: duration_ns of the node in nanoseconds.
            created_at: Optional[datetime]: Timestamp of the span's creation.
            user_metadata: Optional[Dict[str, str]]: Metadata associated with this span.
            id: Optional[UUID4]: ID of the span.
            step_number: Optional[int]: Step number of the span.
        Returns:
        -------
            WorkflowSpan: The created span.
        """
        span = WorkflowSpan(
            input=input,
            redacted_input=redacted_input,
            output=output,
            redacted_output=redacted_output,
            name=name,
            created_at=self._get_child_span_timestamp() if created_at is None else created_at,
            user_metadata=user_metadata,
            tags=tags,
            metrics=Metrics(duration_ns=duration_ns),
            id=id,
            step_number=step_number,
        )
        self.add_child_span_to_parent(span)
        self._parent_stack.append(span)
        return span

    def add_agent_span(
        self,
        input: str,
        redacted_input: Optional[str] = None,
        output: Optional[str] = None,
        redacted_output: Optional[str] = None,
        name: Optional[str] = None,
        duration_ns: Optional[int] = None,
        created_at: Optional[datetime] = None,
        user_metadata: Optional[Dict[str, str]] = None,
        tags: Optional[List[str]] = None,
        agent_type: Optional[AgentType] = None,
        id: Optional[UUID4] = None,
        step_number: Optional[int] = None,
    ) -> AgentSpan:
        """
        Parameters:
        ----------
            input: str: Input to the node.
            redacted_input: Optional[str]: Redacted input to the node.
            output: Optional[str]: Output of the node. This can also be set on conclude().
            redacted_output: Optional[str]: Redacted output of the node.
            name: Optional[str]: Name of the span.
            duration_ns: Optional[int]: duration_ns of the node in nanoseconds.
            created_at: Optional[datetime]: Timestamp of the span's creation.
            user_metadata: Optional[Dict[str, str]]: Metadata associated with this span.
            agent_type: Optional[AgentType]: Agent type of the span.
            step_number: Optional[int]: Step number of the span.
        Returns
        -------
            AgentSpan: The created span.
        """
        span = AgentSpan(
            input=input,
            redacted_input=redacted_input,
            output=output,
            redacted_output=redacted_output,
            name=name,
            created_at=self._get_child_span_timestamp() if created_at is None else created_at,
            user_metadata=user_metadata,
            tags=tags,
            metrics=Metrics(duration_ns=duration_ns),
            agent_type=agent_type,
            id=id,
            step_number=step_number,
        )
        self.add_child_span_to_parent(span)
        self._parent_stack.append(span)
        return span

    def conclude(
        self,
        output: Optional[str] = None,
        redacted_output: Optional[str] = None,
        duration_ns: Optional[int] = None,
        status_code: Optional[int] = None,
    ) -> Optional[StepWithChildSpans]:
        """
        Conclude the current trace or workflow span by setting the output of the current node. In the case of nested
        workflow spans, this will point the workflow back to the parent of the current workflow span.

        Parameters:
        ----------
            output: Optional[str]: Output of the node.
            redacted_output: Optional[str]: Redacted output of the node.
            duration_ns: Optional[int]: duration_ns of the node in nanoseconds.
            status_code: Optional[int]: Status code of the node execution.
        Returns:
        -------
            Optional[StepWithChildSpans]: The parent of the current workflow. None if no parent exists.
        """
        current_parent = self.current_parent()
        if current_parent is None:
            raise ValueError("No existing workflow to conclude.")

        current_parent.output = output or current_parent.output
        if redacted_output is not None:
            current_parent.redacted_output = redacted_output or current_parent.redacted_output
        current_parent.status_code = status_code
        if duration_ns is not None:
            current_parent.metrics.duration_ns = duration_ns

        finished_step = self._parent_stack.pop()
        if self.current_parent() is None and not isinstance(finished_step, Trace):
            raise ValueError("Finished step is not a trace, but has no parent.  Not added to the list of traces.")
        return self.current_parent()



from enum import Enum, unique


@unique
class JobStatus(str, Enum):
    unstarted = "unstarted"  # When the job is created but not yet started.
    in_progress = "in_progress"  # When the job is running.
    processed = "processed"  # When the job is processed and results are in the object store.
    completed = "completed"  # When the job is finished successfully.
    error = "error"  # When the job fails because of a user error.
    failed = "failed"  # When the job fails because of a system error.

    @staticmethod
    def is_incomplete(status: "JobStatus") -> bool:
        return status in [JobStatus.unstarted, JobStatus.in_progress]

    @staticmethod
    def is_failed(status: "JobStatus") -> bool:
        return status in [JobStatus.failed, JobStatus.error]


@unique
class JobName(str, Enum):
    """
    Function name for runners entrypoint.

    For background jobs submitted to runners, the JobName dictate to the service runners which functions should be run.
    See 'BaseVaexService.run()' for more information.
    """

    default = "default"
    inference = "inference"
    lykos_migration = "lykos_migration"
    monitor_run = "monitor_run"
    monitor_scorer = "monitor_scorer"
    prompt_chain_run = "prompt_chain_run"
    prompt_optimization = "prompt_optimization"
    prompt_rater = "prompt_rater"
    prompt_run = "prompt_run"
    prompt_scorer = "prompt_scorer"
    protect_scorer = "protect_scorer"
    registered_scorer_validation = "registered_scorer_validation"
    metric_critique = "metric_critique"
    auto_gen = "auto_gen"
    generated_scorer_validation = "generated_scorer_validation"
    log_record_generated_scorer_validation = "log_record_generated_scorer_validation"
    log_stream_run = "log_stream_run"
    playground_run = "playground_run"
    log_stream_scorer = "log_stream_scorer"
    synthetic_datagen = "synthetic_datagen"
    logstream_insights = "logstream_insights"
    auto_metric_suggestion = "auto_metric_suggestion"



from typing import Dict, List, Optional

from pydantic import TypeAdapter

from galileo_core.constants.request_method import RequestMethod
from galileo_core.constants.routes import Routes
from galileo_core.helpers.logger import logger
from galileo_core.schemas.base_config import GalileoConfig
from galileo_core.schemas.core.integration.anthropic_auth_type import AnthropicAuthenticationType
from galileo_core.schemas.core.integration.azure import DEFAULT_AZURE_API_VERSION, AzureModelDeployment
from galileo_core.schemas.core.integration.azure_auth_type import AzureAuthenticationType
from galileo_core.schemas.core.integration.base import (
    IntegrationResponse,
)
from galileo_core.schemas.core.integration.create_integration import CreateIntegrationModel
from galileo_core.schemas.core.integration.integration_type import LLMIntegration


def list_integrations(config: Optional[GalileoConfig] = None) -> List[IntegrationResponse]:
    """
    Returns all integrations that the user has access to.

    Returns
    -------
    List[IntegrationResponse]
        A list of integrations.
    """
    config = config or GalileoConfig.get()

    logger.debug("Getting integrations...")
    integrations = [
        IntegrationResponse.model_validate(integration)
        for integration in config.api_client.request(RequestMethod.GET, Routes.integrations)
    ]
    logger.debug(f"Got {len(integrations)} integrations.")
    return integrations


def create_or_update_integration(name: str, data: dict, config: Optional[GalileoConfig] = None) -> IntegrationResponse:
    """
    Create or update an integration.

    Parameters
    ----------
    name : str
        A name of the new integration.

    data: dict
        All additional data for the new integration.

    Returns
    -------
    IntegrationResponse
        Response object for the created or updated integration.
    """
    config = config or GalileoConfig.get()

    logger.debug(f"Creating integration with name {name}...")
    adapter: TypeAdapter[CreateIntegrationModel] = TypeAdapter(CreateIntegrationModel)
    validated_data = adapter.validate_python({"name": name, **data})

    response_dict = config.api_client.request(
        RequestMethod.PUT,
        Routes.create_update_integration.format(integration_name=name),
        json=validated_data.model_dump(),
    )
    integration_response = IntegrationResponse.model_validate(response_dict)
    logger.debug(f"Created integration with name {integration_response.name}, ID {integration_response.id}.")
    return integration_response


def create_or_update_openai_integration(
    api_key: str,
    organization_id: Optional[str] = None,
    config: Optional[GalileoConfig] = None,
) -> IntegrationResponse:
    """
    Create or update an OpenAI integration.

    Parameters
    ----------
    api_key : str
        OpenAI API key.
    organization_id : Optional[str], optional
        OpenAI organization ID, by default None.

    Returns
    -------
    IntegrationResponse
        Response object for the created or updated integration.
    """
    config = config or GalileoConfig.get()
    logger.debug("Creating OpenAI integration...")
    data = {"token": api_key, "organization_id": organization_id}
    return create_or_update_integration(LLMIntegration.openai, data, config)


def create_or_update_azure_integration(
    api_key: str,
    endpoint: str,
    proxy: bool = False,
    authentication_type: AzureAuthenticationType = AzureAuthenticationType.api_key,
    authentication_scope: Optional[str] = None,
    headers: Optional[Dict[str, str]] = None,
    api_version: str = DEFAULT_AZURE_API_VERSION,
    azure_deployment: Optional[str] = None,
    available_deployments: Optional[List[AzureModelDeployment]] = None,
    oauth2_token_url: Optional[str] = None,
    custom_header_mapping: Optional[Dict[str, str]] = None,
    config: Optional[GalileoConfig] = None,
) -> IntegrationResponse:
    """
    Create or update an Azure integration.

    Parameters
    ----------
    api_key : str
        Azure API key.
    endpoint : str
        The Azure OpenAI endpoint.
    proxy : bool, optional
        Whether to use a proxy when making requests, by default False.
    authentication_type : AzureAuthenticationType, optional
        Type of authentication to use.
    authentication_scope : Optional[str], optional
        Scope for authentication, if applicable.
    headers : Optional[dict[str, str]], optional
        Additional headers to include in requests.
    api_version : str, optional
        Azure API version to use, by default DEFAULT_AZURE_API_VERSION.
    azure_deployment : Optional[str], optional
        The default deployment name to use.
    available_deployments : Optional[List[AzureModelDeployment]], optional
        Predefined deployments to avoid querying Azure for them.
    oauth2_token_url : Optional[str], optional
        OAuth2 token URL for custom OAuth2 authentication, if applicable.
    custom_header_mapping : Optional[Dict[str, str]], optional
        Custom header mapping for the integration, by default None.

    Returns
    -------
    IntegrationResponse
        Response object for the created or updated integration.
    """
    config = config or GalileoConfig.get()
    logger.debug("Creating Azure integration...")
    data = {
        "token": api_key,
        "endpoint": endpoint,
        "proxy": proxy,
        "authentication_type": authentication_type,
        "authentication_scope": authentication_scope,
        "headers": headers,
        "api_version": api_version,
        "azure_deployment": azure_deployment,
        "available_deployments": available_deployments,
        "oauth2_token_url": oauth2_token_url,
        "custom_header_mapping": custom_header_mapping,
    }
    return create_or_update_integration(LLMIntegration.azure, data, config)


def create_or_update_vertex_ai_integration(token: str, config: Optional[GalileoConfig] = None) -> IntegrationResponse:
    """
    Create or update a Vertex AI integration.
    Parameters
    ----------
    token : str
        Vertex AI token.
    Returns
    -------
    IntegrationResponse
        Response object for the created or updated integration.
    """
    config = config or GalileoConfig.get()
    logger.debug("Creating Vertex AI integration...")
    data = {"token": token}
    return create_or_update_integration(LLMIntegration.vertex_ai, data, config)


def create_or_update_anthropic_integration(
    api_key: str,
    endpoint: Optional[str] = None,
    authentication_type: AnthropicAuthenticationType = AnthropicAuthenticationType.api_key,
    authentication_scope: Optional[str] = None,
    oauth2_token_url: Optional[str] = None,
    custom_header_mapping: Optional[Dict[str, str]] = None,
    config: Optional[GalileoConfig] = None,
) -> IntegrationResponse:
    """
    Create or update an Anthropic integration.

    Parameters
    ----------
    api_key : str
        Anthropic API key.
    authentication_type : AnthropicAuthenticationType
        Type of authentication to use. This should be `AnthropicAuthenticationType.api_key` for API key authentication or `AnthropicAuthenticationType.custom_oauth2` for custom OAuth2 authentication.
    authentication_scope : Optional[str], optional
        Scope for authentication, if applicable.
    oauth2_token_url : Optional[str], optional
        OAuth2 token URL for custom OAuth2 authentication, if applicable.
    custom_header_mapping : Optional[Dict[str, str]], optional
        Custom header mapping for the integration, by default None.

    Returns
    -------
    IntegrationResponse
        Response object for the created or updated integration.
    """
    config = config or GalileoConfig.get()
    logger.debug("Creating Anthropic integration...")
    data = {
        "token": api_key,
        "authentication_type": authentication_type,
        "endpoint": endpoint,
        "authentication_scope": authentication_scope,
        "oauth2_token_url": oauth2_token_url,
        "custom_header_mapping": custom_header_mapping,
    }
    return create_or_update_integration(LLMIntegration.anthropic, data, config)


def create_or_update_mistral_integration(api_key: str, config: Optional[GalileoConfig] = None) -> IntegrationResponse:
    """
    Create or update a Mistral integration.

    Parameters
    ----------
    api_key : str
        Mistral API key.

    Returns
    -------
    IntegrationResponse
        Response object for the created or updated integration.
    """
    config = config or GalileoConfig.get()
    logger.debug("Creating Mistral integration...")
    data = {"token": api_key}
    return create_or_update_integration(LLMIntegration.mistral, data, config)



from enum import Enum
from typing import Dict


# In python, if multiple enum members have the same value, the first one is the "canonical" one.
# Secondary aliases must come after the canonical one.
# Reference: https://docs.python.org/3.10/library/enum.html#duplicating-enum-members-and-values
class ScorerName(str, Enum):
    action_completion_luna = "action_completion_luna"

    action_advancement_luna = "action_advancement_luna"

    agentic_session_success = "agentic_session_success"
    action_completion = "agentic_session_success"

    agentic_workflow_success = "agentic_workflow_success"
    action_advancement = "agentic_workflow_success"

    agent_efficiency = "agent_efficiency"

    agent_flow = "agent_flow"

    bleu = "bleu"

    chunk_attribution_utilization_luna = "chunk_attribution_utilization_luna"

    chunk_attribution_utilization = "chunk_attribution_utilization"

    completeness_luna = "completeness_luna"

    completeness = "completeness"

    context_adherence = "context_adherence"

    context_adherence_luna = "context_adherence_luna"

    context_relevance = "context_relevance"

    context_relevance_luna = "context_relevance_luna"

    conversation_quality = "conversation_quality"

    correctness = "correctness"

    ground_truth_adherence = "ground_truth_adherence"

    input_pii = "input_pii"
    input_pii_gpt = "input_pii_gpt"

    input_sexist = "input_sexist"
    input_sexism = "input_sexist"

    input_sexist_luna = "input_sexist_luna"
    input_sexism_luna = "input_sexist_luna"

    input_tone = "input_tone"
    input_tone_gpt = "input_tone_gpt"

    input_toxicity = "input_toxicity"

    input_toxicity_luna = "input_toxicity_luna"

    instruction_adherence = "instruction_adherence"

    output_pii = "output_pii"
    output_pii_gpt = "output_pii_gpt"

    output_sexist = "output_sexist"
    output_sexism = "output_sexist"

    output_sexist_luna = "output_sexist_luna"
    output_sexism_luna = "output_sexist_luna"

    output_tone = "output_tone"
    output_tone_gpt = "output_tone_gpt"

    output_toxicity = "output_toxicity"

    output_toxicity_luna = "output_toxicity_luna"

    prompt_injection = "prompt_injection"

    prompt_injection_luna = "prompt_injection_luna"

    prompt_perplexity = "prompt_perplexity"

    rouge = "rouge"

    sql_efficiency = "sql_efficiency"

    sql_adherence = "sql_adherence"

    sql_injection = "sql_injection"

    sql_correctness = "sql_correctness"

    tool_error_rate = "tool_error_rate"

    tool_error_rate_luna = "tool_error_rate_luna"

    tool_selection_quality = "tool_selection_quality"

    tool_selection_quality_luna = "tool_selection_quality_luna"

    uncertainty = "uncertainty"

    user_intent_change = "user_intent_change"


LegacyGPTScorerNameMapping: Dict[str, ScorerName] = {
    "agentic_session_success_gpt": ScorerName.agentic_session_success,
    "agentic_workflow_success_gpt": ScorerName.agentic_workflow_success,
    "chunk_attribution_utilization_gpt": ScorerName.chunk_attribution_utilization,
    "completeness_gpt": ScorerName.completeness,
    "context_adherence_gpt": ScorerName.context_adherence,
    "input_sexist_gpt": ScorerName.input_sexist,
    "input_toxicity_gpt": ScorerName.input_toxicity,
    "output_sexist_gpt": ScorerName.output_sexist,
    "output_toxicity_gpt": ScorerName.output_toxicity,
    "prompt_injection_gpt": ScorerName.prompt_injection,
    "tool_error_rate_gpt": ScorerName.tool_error_rate,
    "tool_selection_quality_gpt": ScorerName.tool_selection_quality,
}



from typing import Dict, Optional

from galileo_core.constants.request_method import RequestMethod
from galileo_core.constants.routes import Routes
from galileo_core.helpers.logger import logger
from galileo_core.schemas.base_config import GalileoConfig


def healthcheck(config: Optional[GalileoConfig] = None) -> Dict:
    """
    Check the health of the Galileo API.

    Returns
    -------
    Dict
        Health check response from the Galileo API.
    """
    config = config or GalileoConfig.get()
    logger.debug("Checking the health of the Galileo API...")
    response_dict = config.api_client.request(RequestMethod.GET, Routes.healthcheck)
    logger.debug("Galileo API is reachable.")
    return response_dict



from enum import Enum
from typing import List, Optional, Union

from pydantic import BaseModel, StringConstraints
from typing_extensions import Annotated

FeedbackTagType = Annotated[str, StringConstraints(min_length=1, max_length=255, strip_whitespace=True)]


class FeedbackType(str, Enum):
    like_dislike = "like_dislike"
    star = "star"
    score = "score"
    tags = "tags"
    text = "text"


class FeedbackRatingInfo(BaseModel):
    feedback_type: FeedbackType
    value: Union[bool, int, str, List[FeedbackTagType]]
    explanation: Optional[str]



from typing import List, Literal, Union

from pydantic import UUID4, BaseModel, Field, field_validator

from galileo_core.schemas.core.auth_method import AuthMethod
from galileo_core.schemas.core.user_role import SystemRole, UserRole

PASS_MIN_LENGTH = 8
INVALID_PASS_MSG = (
    f"Password must be {PASS_MIN_LENGTH} characters or "
    "more and have a mix of uppercase, lowercase, "
    "numbers, and special characters"
)


def valid_pass(password: str) -> bool:
    valid_length = len(password) >= PASS_MIN_LENGTH
    upper = any(c.isupper() for c in password)
    lower = any(c.islower() for c in password)
    special = not password.isalnum()
    return valid_length and upper and lower and special


class InviteUsersRequest(BaseModel):
    auth_method: AuthMethod = AuthMethod.email
    emails: List[str] = Field(default_factory=list)
    role: UserRole = UserRole.user
    group_ids: List[UUID4] = []


class User(BaseModel):
    id: UUID4
    email: str
    role: UserRole = UserRole.user


class CreateUserRequest(BaseModel):
    email: str
    auth_method: Literal[AuthMethod.email] = AuthMethod.email
    password: str
    role: UserRole = UserRole.read_only

    @field_validator("password")
    def validate_password(cls, v: str) -> Union[str, None]:
        if not valid_pass(v):
            raise ValueError(INVALID_PASS_MSG)
        return v


class CreateUserResponse(BaseModel):
    id: UUID4
    email: str
    role: Union[UserRole, SystemRole] = UserRole.user


class UpdateUserRoleRequest(BaseModel):
    role: UserRole



from uuid import UUID

from pydantic import UUID4

NULL_UUID_LITERAL = "00000000-0000-0000-0000-000000000000"
NULL_UUID_V4_LITERAL = "00000000-0000-4000-8000-000000000000"

NULL_UUID: UUID = UUID(NULL_UUID_LITERAL)
NULL_UUID_V4: UUID4 = UUID(NULL_UUID_V4_LITERAL)
